from pyramid.security import Allow
from pyramid.security import Authenticated

from pyramid_rpc.jsonrpc import jsonrpc_method
from pyramid_rpc.xmlrpc import xmlrpc_method

ACLS = {
    'hello': [
        (Allow, Authenticated, 'view'),
    ],
}

class Root(dict):

    def __init__(self, request):
        self.request = request
        self.__acl__ = ACLS.get(request.rpc_method)

def basic(request):
    return 'basic'

@jsonrpc_method(method='exc', endpoint='jsonapi')
@xmlrpc_method(method='exc', endpoint='xmlapi')
def exc_view(request):
    raise Exception()

@jsonrpc_method(method='create', endpoint='jsonapi')
@xmlrpc_method(method='create', endpoint='xmlapi')
def create_view(request, a, b):
    return 'create %s %s' % (a, b)

class RPCHandler(object):
    def __init__(self, request):
        self.request = request

    @jsonrpc_method(method='class_hello', endpoint='jsonapi')
    @xmlrpc_method(method='class_hello', endpoint='xmlapi')
    def class_based_hello(self, name):
        return 'hello, ' + name

@jsonrpc_method(method='hello', endpoint='jsonapi')
@xmlrpc_method(method='hello', endpoint='xmlapi')
def basic_hello(request, name):
    return 'hello, %s, stranger' % name

@jsonrpc_method(method='hello', endpoint='secure_jsonapi', permission='view')
@xmlrpc_method(method='hello', endpoint='secure_xmlapi', permission='view')
def auth_hello(request, name):
    return 'hello, %s, friend' % name

def includeme(config):
    config.include('pyramid_rpc.jsonrpc')
    config.include('pyramid_rpc.xmlrpc')

    config.add_jsonrpc_endpoint('jsonapi', '/api/jsonrpc')
    config.add_xmlrpc_endpoint('xmlapi', '/api/xmlrpc')

    config.add_jsonrpc_method(basic, endpoint='jsonapi', method='basic')
    config.add_xmlrpc_method(basic, endpoint='xmlapi', method='basic')

    config.add_jsonrpc_endpoint(
        'secure_jsonapi', '/api/jsonrpc/secure', factory=Root)
    config.add_xmlrpc_endpoint(
        'secure_xmlapi', '/api/xmlrpc/secure', factory=Root)

    config.scan('.')



import unittest
from pyramid import testing

class TestMapplyViewMapper(unittest.TestCase):
    def _makeOne(self, **kw):
        from pyramid_rpc.mapper import MapplyViewMapper
        return MapplyViewMapper(**kw)

    def test___call__isfunc_no_rpc_args_no_attr(self):
        data = '123'
        def view(request):
            return data
        context = testing.DummyResource()
        request = testing.DummyRequest()
        mapper = self._makeOne()
        result = mapper(view)(context, request)
        self.assertEqual(result, data)

    def test___call__isfunc_no_rpc_args_with_attr(self):
        view = lambda *arg: 'wrong'
        data = '123'
        def foo(request):
            return data
        view.foo = foo
        context = testing.DummyResource()
        request = testing.DummyRequest()
        mapper = self._makeOne(attr='foo')
        result = mapper(view)(context, request)
        self.assertEqual(result, data)

    def test___call__isfunc_with_rpc_args(self):
        def view(request, a, b):
            return a, b
        context = testing.DummyResource()
        request = testing.DummyRequest()
        request.rpc_args = ('a', 'b')
        mapper = self._makeOne()
        result = mapper(view)(context, request)
        self.assertEqual(result, ('a', 'b'))

    def test___call__isfunc_with_rpc_args_and_matchdict(self):
        def view(request, a, b, c=1):
            return a, b, c
        context = testing.DummyResource()
        request = testing.DummyRequest()
        request.rpc_args = ('a', 'b')
        request.matchdict = dict(c='2')
        mapper = self._makeOne()
        result = mapper(view)(context, request)
        self.assertEqual(result, ('a', 'b', '2'))

    def test___call__isclass_no_rpc_args_no_attr(self):
        data = '123'
        class view(object):
            def __init__(self, request):
                pass
            def __call__(self):
                return data
        context = testing.DummyResource()
        request = testing.DummyRequest()
        mapper = self._makeOne()
        result = mapper(view)(context, request)
        self.assertEqual(result, data)
        self.assertEqual(request.__view__.__class__, view)

    def test___call__isclass_no_rpc_args_with_attr(self):
        view = lambda *arg: 'wrong'
        data = '123'
        class view(object):
            def __init__(self, request):
                pass
            def index(self):
                return data
        context = testing.DummyResource()
        request = testing.DummyRequest()
        mapper = self._makeOne(attr='index')
        result = mapper(view)(context, request)
        self.assertEqual(result, data)
        self.assertEqual(request.__view__.__class__, view)

    def test___call__isclass_with_rpc_args(self):
        class view(object):
            def __init__(self, request):
                pass
            def __call__(self, a, b):
                return a, b
        context = testing.DummyResource()
        request = testing.DummyRequest()
        request.rpc_args = ('a', 'b')
        mapper = self._makeOne()
        result = mapper(view)(context, request)
        self.assertEqual(result, ('a', 'b'))

    def test___call__isclass_with_rpc_args_and_matchdict(self):
        class view(object):
            def __init__(self, request):
                pass
            def __call__(self, a, b, c=1):
                return a, b, c
        context = testing.DummyResource()
        request = testing.DummyRequest()
        request.rpc_args = ('a', 'b')
        request.matchdict = dict(c='2')
        mapper = self._makeOne()
        result = mapper(view)(context, request)
        self.assertEqual(result, ('a', 'b', '2'))

    def test_mapply_toomanyargs(self):
        def aview(one, two): pass
        mapper = self._makeOne()
        self.assertRaises(TypeError, mapper.mapply, aview, (1, 2, 3), {})

    def test_mapply_all_kwarg_arg_omitted(self):
        def aview(one, two): pass
        mapper = self._makeOne()
        self.assertRaises(TypeError, mapper.mapply, aview, (), dict(one=1))

    def test_mapply_all_kwarg_arg_omitted_with_default(self):
        def aview(one, two=2):
            return one, two
        mapper = self._makeOne()
        result = mapper.mapply(aview, (), dict(one=1))
        self.assertEqual(result, (1, 2))

    def test_mapply_all_kwarg(self):
        def aview(one, two):
            return one, two
        mapper = self._makeOne()
        result = mapper.mapply(aview, (), dict(one=1, two=2))
        self.assertEqual(result, (1, 2))

    def test_mapply_instmethod(self):
        mapper = self._makeOne()
        result = mapper.mapply(self._aview, ('a',), {})
        self.assertEqual(result, 'a')

    def test_mapply_inst(self):
        class Foo(object):
            def __call__(self, a):
                return a
        foo = Foo()
        mapper = self._makeOne()
        result = mapper.mapply(foo, ('a',), {})
        self.assertEqual(result, 'a')

    def _aview(self, a):
        return a



import json
import logging

import venusian
from pyramid.exceptions import ConfigurationError
from pyramid.httpexceptions import HTTPForbidden
from pyramid.httpexceptions import HTTPNotFound
from pyramid.renderers import null_renderer
from pyramid.renderers import render
from pyramid.request import Request
from pyramid.response import Response
from pyramid.security import NO_PERMISSION_REQUIRED

from pyramid_rpc.compat import is_nonstr_iter
from pyramid_rpc.mapper import MapplyViewMapper
from pyramid_rpc.mapper import ViewMapperArgsInvalid
from pyramid_rpc.util import combine


log = logging.getLogger(__name__)

DEFAULT_RENDERER = 'pyramid_rpc:jsonrpc'

_marker = object()


class JsonRpcError(Exception):
    code = -32603 # sane default
    message = 'internal error' # sane default
    data = None

    def __init__(self, code=None, message=None, data=None):
        if code is not None:
            self.code = code
        if message is not None:
            self.message = message
        if data is not None:
            self.data = data

    def as_dict(self):
        """Return a dictionary representation of this object for
        serialization in a JSON-RPC response."""
        error = dict(code=self.code,
                     message=self.message)
        if self.data is not None:
            error['data'] = self.data
        return error


class JsonRpcParseError(JsonRpcError):
    code = -32700
    message = 'parse error'


class JsonRpcRequestInvalid(JsonRpcError):
    code = -32600
    message = 'invalid request'


class JsonRpcMethodNotFound(JsonRpcError):
    code = -32601
    message = 'method not found'


class JsonRpcParamsInvalid(JsonRpcError):
    code = -32602
    message = 'invalid params'


class JsonRpcInternalError(JsonRpcError):
    code = -32603
    message = 'internal error'


def make_error_response(request, error, id=None):
    """ Marshal a Python Exception into a ``Response`` object with a
    body that is a JSON string suitable for use as a JSON-RPC response
    with a content-type of ``application/json`` and return the response.

    """
    # we may need to render a parse error, at which point we don't know
    # much about the request
    renderer = getattr(request, 'rpc_renderer', DEFAULT_RENDERER)
    out = {
        'jsonrpc': '2.0',
        'id': id,
        'error': error.as_dict(),
    }
    body = render(renderer, out, request=request).encode('utf-8')

    response = Response(body, charset='utf-8')
    response.content_type = 'application/json'
    return response


def exception_view(exc, request):
    rpc_id = getattr(request, 'rpc_id', None)
    if isinstance(exc, JsonRpcError):
        fault = exc
        log.debug('json-rpc error rpc_id:%s "%s"',
                  rpc_id, exc.message)
    elif isinstance(exc, HTTPNotFound):
        fault = JsonRpcMethodNotFound()
        log.debug('json-rpc method not found rpc_id:%s "%s"',
                  rpc_id, request.rpc_method)
    elif isinstance(exc, HTTPForbidden):
        fault = JsonRpcRequestInvalid()
        log.debug('json-rpc method forbidden rpc_id:%s "%s"',
                  rpc_id, request.rpc_method)
    elif isinstance(exc, ViewMapperArgsInvalid):
        fault = JsonRpcParamsInvalid()
        log.debug('json-rpc invalid method params')
    else:
        fault = JsonRpcInternalError()
        log.exception('json-rpc exception rpc_id:%s "%s"', rpc_id, exc)

    return make_error_response(request, fault, rpc_id)


def make_response(request, result):
    rpc_id = getattr(request, 'rpc_id', None)
    response = request.response

    # store content_type before render is called
    ct = response.content_type

    out = {
        'jsonrpc': '2.0',
        'id': rpc_id,
        'result': result,
    } if request.rpc_id is not None else ''
    response.body = render(
        request.rpc_renderer, out, request=request
    ).encode(response.charset)

    if ct == response.default_content_type:
        response.content_type = 'application/json'

    return response


def _render(value, system):
    return json.dumps(value)


def jsonrpc_renderer(info):
    return _render


class jsonrpc_view(object):
    """ Decorator that wraps a view and converts the result into a valid
    JSON-RPC Response object.

    """
    def __init__(self, renderer=DEFAULT_RENDERER):
        self.renderer = renderer

    def __call__(self, wrapped):
        def wrapper(context, request):
            request.rpc_renderer = self.renderer
            result = wrapped(context, request)
            if not request.is_response(result):
                result = make_response(request, result)
            return result
        return wrapper


def parse_request_GET(request):
    """ Parse JSON-RPC parameters from the request query string."""
    args = request.GET.get('params')
    if args is not None:
        try:
            request.rpc_args = json.loads(args)
        except ValueError:
            raise JsonRpcParseError
    else:
        request.rpc_args = ()

    request.rpc_method = request.GET.get('method')
    request.rpc_id = request.GET.get('id')
    request.rpc_version = request.GET.get('jsonrpc')


def parse_request_POST(request):
    """ Parse JSON-RPC parameters from the request body."""
    try:
        body = request.json_body
    except ValueError:
        raise JsonRpcParseError

    try:
        batched = body[:]
    except TypeError:
        batched = None

    if batched is not None:
        request.batched_rpc_requests = batched
    else:
        request.rpc_id = body.get('id')
        request.rpc_args = body.get('params', ())
        request.rpc_method = body.get('method')
        request.rpc_version = body.get('jsonrpc')


def setup_request(endpoint, request):
    """ Parse a JSON-RPC request body."""
    if request.method == 'GET':
        parse_request_GET(request)
    elif request.method == 'POST':
        parse_request_POST(request)
    else:
        log.debug('unsupported request method "%s"', request.method)
        raise JsonRpcRequestInvalid

    if hasattr(request, 'batched_rpc_requests'):
        log.debug('handling batched rpc request')
        # the checks below will look at the subrequests
        return

    if request.rpc_version != '2.0':
        log.debug('id:%s invalid rpc version %s',
                  request.rpc_id, request.rpc_version)
        raise JsonRpcRequestInvalid

    if request.rpc_method is None:
        log.debug('id:%s invalid rpc method', request.rpc_id)
        raise JsonRpcRequestInvalid

    log.debug('handling id:%s method:%s',
              request.rpc_id, request.rpc_method)


class EndpointPredicate(object):
    def __init__(self, val, config):
        self.val = val

    def text(self):
        return 'jsonrpc endpoint = %s' % self.val

    phash = text

    def __call__(self, info, request):
        if self.val:
            # find the endpoint info
            key = info['route'].name
            endpoint = request.registry.jsonrpc_endpoints[key]

            # potentially setup either rpc v1 or v2 from the parsed body
            setup_request(endpoint, request)

            # update request with endpoint information
            request.rpc_endpoint = endpoint

            # Always return True so that even if it isn't a valid RPC it
            # will fall through to the notfound_view which will still
            # return a valid JSON-RPC response.
            return True


class MethodPredicate(object):
    def __init__(self, val, config):
        self.method = val

    def text(self):
        return 'jsonrpc method = %s' % self.method

    phash = text

    def __call__(self, context, request):
        return getattr(request, 'rpc_method') == self.method


class BatchedRequestPredicate(object):
    def __init__(self, val, config):
        self.val = val

    def text(self):
        return 'jsonrpc batched request = %s' % self.val

    phash = text

    def __call__(self, context, request):
        if self.val:
            return hasattr(request, 'batched_rpc_requests')


def batched_request_view(request):
    json_response = []
    response = request.response
    for rpc_request in request.batched_rpc_requests:
        body = json.dumps(rpc_request).encode(request.charset)
        subrequest = Request.blank(path=request.path,
                                   environ=request.environ,
                                   base_url=request.application_url,
                                   headers=request.headers,
                                   POST=body,
                                   charset=request.charset)
        subresponse = request.invoke_subrequest(subrequest, use_tweens=True)
        if subresponse.json_body != '':
            json_response.append(subresponse.json_body)
    if json_response:
        # use charset and content-type from last subresponse
        response.charset = subresponse.charset
        response.content_type = subresponse.content_type
        # will automatically be encoded
        response.json_body = json_response
    else:
        # if we would send an empty list, instead send nothing
        # per JSON-RPC: http://www.jsonrpc.org/specification#batch
        response.content_type = 'text/plain'
        response.body = b''
    return response


class Endpoint(object):
    def __init__(self, name, default_mapper, default_renderer):
        self.name = name
        self.default_mapper = default_mapper
        self.default_renderer = default_renderer


def add_jsonrpc_endpoint(config, name, *args, **kw):
    """Add an endpoint for handling JSON-RPC.

    ``name``

        The name of the endpoint.

    ``default_mapper``

        A default view mapper that will be passed as the ``mapper``
        argument to each of the endpoint's methods.

    ``default_renderer``

        A default renderer that will be passed as the ``renderer``
        argument to each of the endpoint's methods. This should be the
        string name of the renderer, registered via
        :meth:`pyramid.config.Configurator.add_renderer`.

    A JSON-RPC method also accepts all of the arguments supplied to
    :meth:`pyramid.config.Configurator.add_route`.

    """
    default_mapper = kw.pop('default_mapper', MapplyViewMapper)
    default_renderer = kw.pop('default_renderer', DEFAULT_RENDERER)

    endpoint = Endpoint(
        name,
        default_mapper=default_mapper,
        default_renderer=default_renderer,
    )

    config.registry.jsonrpc_endpoints[name] = endpoint

    kw['jsonrpc_endpoint'] = True
    config.add_route(name, *args, **kw)

    kw = {}
    kw['jsonrpc_batched'] = True
    kw['renderer'] = null_renderer
    config.add_view(batched_request_view, route_name=name,
                    permission=NO_PERMISSION_REQUIRED, **kw)
    config.add_view(exception_view, route_name=name, context=Exception,
                    permission=NO_PERMISSION_REQUIRED)


def add_jsonrpc_method(config, view, **kw):
    """Add a method to a JSON-RPC endpoint.

    ``endpoint``

        The name of the endpoint.

    ``method``

        The name of the method.

    A JSON-RPC method also accepts all of the arguments supplied to
    :meth:`pyramid.config.Configurator.add_view`.

    A view mapper is registered by default which will match the
    ``request.rpc_args`` to parameters on the view. To override this
    behavior simply set the ``mapper`` argument to None or another
    view mapper.

    .. note::

       An endpoint **must** be defined before methods may be added.

    """
    endpoint_name = kw.pop('endpoint', kw.pop('route_name', None))
    if endpoint_name is None:
        raise ConfigurationError(
            'Cannot register a JSON-RPC endpoint without specifying the '
            'name of the endpoint.')

    endpoint = config.registry.jsonrpc_endpoints.get(endpoint_name)
    if endpoint is None:
        raise ConfigurationError(
            'Could not find an endpoint with the name "%s".' % endpoint_name)

    # pop the method name
    method = kw.pop('method', None)
    if method is None:
        raise ConfigurationError(
            'Cannot register a JSON-RPC method without specifying the '
            '"method"')

    mapper = kw.pop('mapper', _marker)
    if mapper is _marker:
        # only override mapper if not supplied
        mapper = endpoint.default_mapper
    kw['mapper'] = mapper

    renderer = kw.pop('renderer', None)
    if renderer is None:
        renderer = endpoint.default_renderer
    kw['renderer'] = null_renderer

    kw['jsonrpc_method'] = method

    rpc_decorator = jsonrpc_view(renderer)
    decorator = kw.get('decorator', None)
    if decorator is None:
        decorator = rpc_decorator
    else:
        if not is_nonstr_iter(decorator):
            decorator = (decorator,)
        # we want to apply the view_wrapper first, then the other decorators
        # and combine() reverses the order, so ours goes last
        decorators = list(decorator) + [rpc_decorator]
        decorator = combine(*decorators)
    kw['decorator'] = decorator

    config.add_view(view, route_name=endpoint_name, **kw)


class jsonrpc_method(object):
    """This decorator may be used with pyramid view callables to enable
    them to respond to JSON-RPC method calls.

    If ``method`` is not supplied, then the callable name will be used
    for the method name.

    ``_depth`` may be specified when wrapping ``jsonrpc_method`` in another
    decorator. The value should reflect how many stack frames are between
    the wrapped target and ``jsonrpc_method``. Thus a decorator one level deep
    would pass in ``_depth=1``.

    This is the lazy analog to the
    :func:`~pyramid_rpc.jsonrpc.add_jsonrpc_method`` and accepts all of
    the same arguments.

    """
    def __init__(self, method=None, **kw):
        self.method = method
        self.kw = kw

    def __call__(self, wrapped):
        kw = self.kw.copy()
        kw['method'] = self.method or wrapped.__name__
        depth = kw.pop('_depth', 0)

        def callback(context, name, ob):
            config = context.config.with_package(info.module)
            config.add_jsonrpc_method(view=ob, **kw)

        info = venusian.attach(wrapped, callback, category='pyramid',
                               depth=depth + 1)
        if info.scope == 'class':
            # ensure that attr is set if decorating a class method
            kw.setdefault('attr', wrapped.__name__)

        kw['_info'] = info.codeinfo  # fbo action_method
        return wrapped


def includeme(config):
    """ Set up standard configurator registrations.  Use via:

    .. code-block:: python

       config = Configurator()
       config.include('pyramid_rpc.jsonrpc')

    Once this function has been invoked, two new directives will be
    available on the configurator:

    - ``add_jsonrpc_endpoint``: Add an endpoint for handling JSON-RPC.

    - ``add_jsonrpc_method``: Add a method to a JSON-RPC endpoint.

    """
    if not hasattr(config.registry, 'jsonrpc_endpoints'):
        config.registry.jsonrpc_endpoints = {}

    config.add_view_predicate('jsonrpc_method', MethodPredicate)
    config.add_view_predicate('jsonrpc_batched', BatchedRequestPredicate)
    config.add_route_predicate('jsonrpc_endpoint', EndpointPredicate)

    config.add_renderer(DEFAULT_RENDERER, jsonrpc_renderer)
    config.add_directive('add_jsonrpc_endpoint', add_jsonrpc_endpoint)
    config.add_directive('add_jsonrpc_method', add_jsonrpc_method)
    config.add_view(exception_view, context=JsonRpcError,
                    permission=NO_PERMISSION_REQUIRED)



import unittest

from pyramid import testing

try:
    import pyamf
    from pyamf import remoting, util
except ImportError: # pragma: nocover
    from nose import SkipTest
    raise SkipTest

class DummyLogging(object):
    def exception(self, *args):
        pass

    def debug(self, *args):
        pass

class TestAmfViewGateway(unittest.TestCase):
    def _makeOne(self, **kwargs):
        from pyramid_rpc.amfgateway import PyramidGateway
        return PyramidGateway(logger=DummyLogging(), **kwargs)

    def _makeRequestBody(self, service, body, raw=False):
        if not raw:
            body = [body]

        e = remoting.Envelope(pyamf.AMF3)
        e['/1'] = remoting.Request(service, body=body)
        return remoting.encode(e)

    def doRequest(self, body, gw_instance, request_method='POST', **kwargs):
        request = testing.DummyRequest(**kwargs)
        request.body = body
        request.method = request_method
        return gw_instance(request)
    
    def test_request_method(self):
        gw = self._makeOne()
        response = self.doRequest(None, gw, 'GET')
        self.assertEqual(response.status_int, 405)

    def test_bad_request(self):
        gw = self._makeOne()
        body = util.BufferedByteStream()
        body.write('Bad request')
        body.seek(0, 0)
        
        response = self.doRequest(body, gw)
        self.assertEqual(response.status_int, 400)

    def test_unknown_request(self):
        gw = self._makeOne()
        body = self._makeRequestBody('test.test', [], raw=True)
        
        response = self.doRequest(body, gw)
        
        envelope = remoting.decode(response.body)
        message = envelope['/1']

        self.assertEqual(message.status, remoting.STATUS_ERROR)
        body = message.body

        self.assertTrue(isinstance(body, remoting.ErrorFault))
        self.assertEqual(body.code, 'Service.ResourceNotFound')

    def test_eof_decode(self):
        gw = self._makeOne()
        gw.debug = True
        body = util.BufferedByteStream()
        body.seek(0, 0)
        response = self.doRequest(body, gw)
        assert 'The request body was unable to be successfully decoded' in response.detail
        assert 'Traceback' in response.detail
        self.assertEqual(response.status_int, 400)

    def _raiseException(self, e, *args, **kwargs):
        raise e()

    def _restoreEncode(self):
        remoting.encode = self.old_method
    
    def _restoreDecode(self):
        remoting.decode = self.old_method

    def test_really_bad_decode(self):
        self.old_method = remoting.decode
        remoting.decode = lambda *args, **kwargs: self._raiseException(Exception, *args, **kwargs)
        try:
            gw = self._makeOne()
            gw.debug = True
            request = util.BufferedByteStream()
            response = self.doRequest(request, gw)

            self.assertEqual(response.status_int, 500)
            assert 'Traceback' in response.detail
        finally:
            self._restoreDecode()

    def test_expected_exceptions_decode(self):
        self.old_method = remoting.decode
        try:
            gw = self._makeOne()
            request = util.BufferedByteStream()

            for x in (KeyboardInterrupt, SystemExit):
                remoting.decode = lambda *args, **kwargs: self._raiseException(x, *args, **kwargs)

                self.assertRaises(x, self.doRequest, request, gw)
        finally:
            self._restoreDecode()
    
    def test_expected_exception_response(self):
        gw = self._makeOne()
        request = self._makeRequestBody('echo', 'hello')

        gw.getResponse = lambda *args, **kwargs: self._raiseException(KeyboardInterrupt, *args, **kwargs)
        self.assertRaises(KeyboardInterrupt, self.doRequest, request, gw)        

    def test_expected_exception_encode(self):
        self.old_method = remoting.encode
        try:
            gw = self._makeOne()
            gw.debug = True
            request = self._makeRequestBody('echo', 'hello')
            remoting.encode = lambda *args, **kwargs: self._raiseException(KeyboardInterrupt, *args, **kwargs)
            response = self.doRequest(request, gw)
            self.assertEqual(response.status_int, 500)
            assert 'unable to be encoded' in response.detail
            assert 'Traceback' in response.detail
        finally:
            self._restoreEncode()
        
    def test_expose_request(self):
        gw = self._makeOne()
        gw.expose_request = True
        
        executed = []
        
        def echo(http_request, data):
            assert hasattr(http_request, 'amf_request')
            request = http_request.amf_request

            self.assertTrue(isinstance(request, remoting.Request))

            self.assertEqual(request.target, 'echo')
            self.assertEqual(request.body, ['hello'])
            executed.append(True)

        gw.addService(echo)
        self.doRequest(self._makeRequestBody('echo', 'hello'), gw)
        assert len(executed) > 0

    def test_expected_exception_echo(self):
        gw = self._makeOne()
        gw.expose_request = True
        gw.debug = True
        gw.getResponse = lambda *args, **kwargs: self._raiseException(Exception, *args, **kwargs)
        response = self.doRequest(self._makeRequestBody('echo', 'hello'), gw)
        self.assertEqual(response.status_int, 500)
        assert 'Traceback' in response.detail

    def test_timezone(self):
        import datetime

        gw = self._makeOne()
        gw.expose_request = False
        td = datetime.timedelta(hours=-5)
        now = datetime.datetime.utcnow()

        def echo(d):
            self.assertEqual(d, now + td)
            return d

        gw.addService(echo)
        gw.timezone_offset = -18000

        response = self.doRequest(self._makeRequestBody('echo', now), gw)
        envelope = remoting.decode(response.body)
        message = envelope['/1']
        self.assertEqual(message.body, now)



import sys
import unittest

from pyramid import testing

from webtest import TestApp

from pyramid_rpc.compat import PY3
from pyramid_rpc.compat import xmlrpclib

class TestXMLRPCIntegration(unittest.TestCase):

    def setUp(self):
        self.config = testing.setUp()

    def tearDown(self):
        testing.tearDown()

    def _callFUT(self, app, method, params):
        if PY3: # pragma: no cover
            xml = xmlrpclib.dumps(params, methodname=method).encode('utf-8')
        else:
            xml = xmlrpclib.dumps(params, methodname=method)
        resp = app.post('/api/xmlrpc', content_type='text/xml', params=xml)
        self.assertEqual(resp.status_int, 200)
        self.assertEqual(resp.content_type, 'text/xml')
        return xmlrpclib.loads(resp.body)[0][0]

    def test_add_xmlrpc_method_with_undefined_endpoint(self):
        from pyramid.exceptions import ConfigurationError
        config = self.config
        config.include('pyramid_rpc.xmlrpc')
        self.assertRaises(ConfigurationError,
                          config.add_xmlrpc_method,
                          lambda r: None, endpoint='rpc', method='dummy')

    def test_add_xmlrpc_method_with_missing_endpoint_param(self):
        from pyramid.exceptions import ConfigurationError
        config = self.config
        config.include('pyramid_rpc.xmlrpc')
        config.add_xmlrpc_endpoint('rpc', '/api/xmlrpc')
        self.assertRaises(ConfigurationError,
                          config.add_xmlrpc_method,
                          lambda r: None, method='dummy')

    def test_add_xmlrpc_method_with_no_method(self):
        from pyramid.exceptions import ConfigurationError
        config = self.config
        config.include('pyramid_rpc.xmlrpc')
        config.add_xmlrpc_endpoint('rpc', '/api/xmlrpc')
        self.assertRaises(ConfigurationError,
                          config.add_xmlrpc_method,
                          lambda r: None, endpoint='rpc')

    def test_it(self):
        def view(request, a, b):
            return {'a': a, 'b': b}
        config = self.config
        config.include('pyramid_rpc.xmlrpc')
        config.add_xmlrpc_endpoint('rpc', '/api/xmlrpc')
        config.add_xmlrpc_method(view, endpoint='rpc', method='dummy')
        app = config.make_wsgi_app()
        app = TestApp(app)
        resp = self._callFUT(app, 'dummy', (2, 3))
        self.assertEqual(resp, {'a': 2, 'b': 3})

    def test_it_with_no_mapper(self):
        def view(request):
            return request.rpc_args[0]
        config = self.config
        config.include('pyramid_rpc.xmlrpc')
        config.add_xmlrpc_endpoint('rpc', '/api/xmlrpc')
        config.add_xmlrpc_method(view, endpoint='rpc', method='dummy',
                                  mapper=None)
        app = config.make_wsgi_app()
        app = TestApp(app)
        resp = self._callFUT(app, 'dummy', (2, 3))
        self.assertEqual(resp, 2)

    def test_it_with_multiple_methods(self):
        def view(request, a, b):
            return a
        config = self.config
        config.include('pyramid_rpc.xmlrpc')
        config.add_xmlrpc_endpoint('rpc', '/api/xmlrpc')
        config.add_xmlrpc_method(view, endpoint='rpc', method='dummy')
        config.add_xmlrpc_method(lambda r: 'fail',
                                  endpoint='rpc', method='dummy2')
        app = config.make_wsgi_app()
        app = TestApp(app)
        resp = self._callFUT(app, 'dummy', (2, 3))
        self.assertEqual(resp, 2)

    def test_it_with_no_method(self):
        config = self.config
        config.include('pyramid_rpc.xmlrpc')
        config.add_xmlrpc_endpoint('rpc', '/api/xmlrpc')
        app = config.make_wsgi_app()
        app = TestApp(app)
        try:
            self._callFUT(app, None, (2, 3))
        except xmlrpclib.Fault:
            exc = sys.exc_info()[1] # 2.5 compat
            self.assertEqual(exc.faultCode, -32601)
        else: # pragma: no cover
            raise AssertionError

    def test_it_with_no_params(self):
        def view(request):
            self.assertEqual(request.rpc_args, ())
            return 'no params'
        config = self.config
        config.include('pyramid_rpc.xmlrpc')
        config.add_xmlrpc_endpoint('rpc', '/api/xmlrpc')
        config.add_xmlrpc_method(view, endpoint='rpc', method='dummy')
        app = config.make_wsgi_app()
        app = TestApp(app)
        resp = self._callFUT(app, 'dummy', ())
        self.assertEqual(resp, 'no params')

    def test_it_with_invalid_method(self):
        config = self.config
        config.include('pyramid_rpc.xmlrpc')
        config.add_xmlrpc_endpoint('rpc', '/api/xmlrpc')
        app = config.make_wsgi_app()
        app = TestApp(app)
        try:
            self._callFUT(app, 'dummy', (2, 3))
        except xmlrpclib.Fault:
            exc = sys.exc_info()[1] # 2.5 compat
            self.assertEqual(exc.faultCode, -32601)
        else: # pragma: no cover
            raise AssertionError

    def test_it_with_invalid_body(self):
        config = self.config
        config.include('pyramid_rpc.xmlrpc')
        config.add_xmlrpc_endpoint('rpc', '/api/xmlrpc')
        app = config.make_wsgi_app()
        app = TestApp(app)
        resp = app.post('/api/xmlrpc', content_type='text/xml',
                        params='<')
        try:
            xmlrpclib.loads(resp.body)
        except xmlrpclib.Fault:
            exc = sys.exc_info()[1] # 2.5 compat
            self.assertEqual(exc.faultCode, -32700)
        else: # pragma: no cover
            raise AssertionError

    def test_it_with_general_exception(self):
        def view(request, a, b):
            raise Exception
        config = self.config
        config.include('pyramid_rpc.xmlrpc')
        config.add_xmlrpc_endpoint('rpc', '/api/xmlrpc')
        config.add_xmlrpc_method(view, endpoint='rpc', method='dummy')
        app = config.make_wsgi_app()
        app = TestApp(app)
        try:
            self._callFUT(app, 'dummy', (2, 3))
        except xmlrpclib.Fault:
            exc = sys.exc_info()[1] # 2.5 compat
            self.assertEqual(exc.faultCode, -32500)
        else: # pragma: no cover
            raise AssertionError

    def test_it_with_cls_view(self):
        class view(object):
            def __init__(self, request):
                self.request = request

            def foo(self, a, b):
                return [a, b]
        config = self.config
        config.include('pyramid_rpc.xmlrpc')
        config.add_xmlrpc_endpoint('rpc', '/api/xmlrpc')
        config.add_xmlrpc_method(view, endpoint='rpc', method='dummy',
                                  attr='foo')
        app = config.make_wsgi_app()
        app = TestApp(app)
        resp = self._callFUT(app, 'dummy', (2, 3))
        self.assertEqual(resp, [2, 3])

    def test_it_with_default_args(self):
        def view(request, a, b, c='bar'):
            return [a, b, c]
        config = self.config
        config.include('pyramid_rpc.xmlrpc')
        config.add_xmlrpc_endpoint('rpc', '/api/xmlrpc')
        config.add_xmlrpc_method(view, endpoint='rpc', method='dummy')
        app = config.make_wsgi_app()
        app = TestApp(app)
        resp = self._callFUT(app, 'dummy', (2, 3))
        self.assertEqual(resp, [2, 3, 'bar'])

    def test_it_with_missing_args(self):
        config = self.config
        config.include('pyramid_rpc.xmlrpc')
        config.add_xmlrpc_endpoint('rpc', '/api/xmlrpc')
        config.add_xmlrpc_method(lambda r, a, b: a,
                                  endpoint='rpc', method='dummy')
        app = config.make_wsgi_app()
        app = TestApp(app)
        try:
            self._callFUT(app, 'dummy', (2,))
        except xmlrpclib.Fault:
            exc = sys.exc_info()[1] # 2.5 compat
            self.assertEqual(exc.faultCode, -32602)
        else: # pragma: no cover
            raise AssertionError

    def test_it_with_too_many_args(self):
        config = self.config
        config.include('pyramid_rpc.xmlrpc')
        config.add_xmlrpc_endpoint('rpc', '/api/xmlrpc')
        config.add_xmlrpc_method(lambda r, a, b: a,
                                  endpoint='rpc', method='dummy')
        app = config.make_wsgi_app()
        app = TestApp(app)
        try:
            self._callFUT(app, 'dummy', (2, 3, 4))
        except xmlrpclib.Fault:
            exc = sys.exc_info()[1] # 2.5 compat
            self.assertEqual(exc.faultCode, -32602)
        else: # pragma: no cover
            raise AssertionError

    def test_it_with_decorator(self):
        def view(request):
            return 'foo'
        config = self.config
        config.include('pyramid_rpc.xmlrpc')
        config.add_xmlrpc_endpoint('rpc', '/api/xmlrpc')
        dummy_decorator = DummyDecorator()
        config.add_xmlrpc_method(view, endpoint='rpc', method='dummy',
                                 decorator=dummy_decorator)
        app = config.make_wsgi_app()
        app = TestApp(app)
        resp = self._callFUT(app, 'dummy', ())
        self.assertEqual(resp, 'foo')
        self.assertTrue(dummy_decorator.called)

    def test_it_with_default_mapper(self):
        def view(request):
            return request.rpc_args
        config = self.config
        config.include('pyramid_rpc.xmlrpc')
        config.add_xmlrpc_endpoint('rpc', '/api/xmlrpc', default_mapper=None)
        config.add_xmlrpc_method(view, endpoint='rpc', method='dummy')
        app = config.make_wsgi_app()
        app = TestApp(app)
        resp = self._callFUT(app, 'dummy', ('a', 'b', 'c'))
        self.assertEqual(resp, ['a', 'b', 'c'])

    def test_override_default_mapper(self):
        from pyramid_rpc.mapper import MapplyViewMapper
        def view(request, a, b, c):
            return (a, b, c)
        config = self.config
        config.include('pyramid_rpc.xmlrpc')
        config.add_xmlrpc_endpoint('rpc', '/api/xmlrpc', default_mapper=None)
        config.add_xmlrpc_method(view, endpoint='rpc', method='dummy',
                                 mapper=MapplyViewMapper)
        app = config.make_wsgi_app()
        app = TestApp(app)
        resp = self._callFUT(app, 'dummy', ('a', 'b', 'c'))
        self.assertEqual(resp, ['a', 'b', 'c'])

    def test_it_with_default_renderer(self):
        def view(request, a, b, c):
            return 'bar'
        config = self.config
        config.include('pyramid_rpc.xmlrpc')
        dummy_renderer = DummyRenderer(('foo',))
        config.add_renderer('xmlrpc', dummy_renderer)
        config.add_xmlrpc_endpoint('rpc', '/api/xmlrpc',
                                   default_renderer='xmlrpc')
        config.add_xmlrpc_method(view, endpoint='rpc', method='dummy')
        app = config.make_wsgi_app()
        app = TestApp(app)
        resp = self._callFUT(app, 'dummy', ('a', 'b', 'c'))
        self.assertEqual(resp, 'foo')
        self.assertEqual(dummy_renderer.called, True)

    def test_override_default_renderer(self):
        def view(request):
            return 'bar'
        config = self.config
        config.include('pyramid_rpc.xmlrpc')
        dummy_renderer = DummyRenderer(('foo',))
        dummy_renderer2 = DummyRenderer(('baz',))
        config.add_renderer('xmlrpc', dummy_renderer)
        config.add_renderer('xmlrpc2', dummy_renderer2)
        config.add_xmlrpc_endpoint('rpc', '/api/xmlrpc',
                                   default_renderer='xmlrpc')
        config.add_xmlrpc_method(view, endpoint='rpc', method='dummy',
                                 renderer='xmlrpc2')
        app = config.make_wsgi_app()
        app = TestApp(app)
        resp = self._callFUT(app, 'dummy', ())
        self.assertEqual(resp, 'baz')
        self.assertEqual(dummy_renderer.called, False)
        self.assertEqual(dummy_renderer2.called, True)

    def test_nonascii_request(self):
        def view(request, a):
            return a
        config = self.config
        config.include('pyramid_rpc.xmlrpc')
        config.add_xmlrpc_endpoint('rpc', '/api/xmlrpc')
        config.add_xmlrpc_method(view, endpoint='rpc', method='dummy')
        app = config.make_wsgi_app()
        app = TestApp(app)
        val = b'S\xc3\xa9bastien'.decode('utf-8')
        resp = self._callFUT(app, 'dummy', (val,))
        self.assertEqual(resp, val)


class DummyDecorator(object):
    called = False

    def __call__(self, view):
        def wrapper(context, request):
            self.called = True
            return view(context, request)
        return wrapper

class DummyRenderer(object):
    called = False

    def __init__(self, result):
        self.result = result

    def __call__(self, info):
        def _render(value, system):
            self.called = True
            system['request'].response.content_type = 'text/xml'
            return xmlrpclib.dumps(self.result, methodresponse=True)
        return _render



import sys
import types

py_version = sys.version_info[:2]
PY3 = py_version[0] == 3


if PY3: # pragma: no cover
    string_types = str,
    integer_types = int,
    class_types = type,
    text_type = str
    binary_type = bytes
    long = int
else:
    string_types = basestring,
    integer_types = (int, long)
    class_types = (type, types.ClassType)
    text_type = unicode
    binary_type = str
    long = long


if PY3: # pragma: no cover
    import xmlrpc.client as xmlrpclib
else:
    import xmlrpclib


if PY3: # pragma: no cover
    def is_nonstr_iter(v):
        if isinstance(v, str):
            return False
        return hasattr(v, '__iter__')
else:
    def is_nonstr_iter(v):
        return hasattr(v, '__iter__')



import json
import unittest

from webtest import TestApp

from pyramid_rpc.compat import xmlrpclib

class DummyAuthenticationPolicy(object):
    userid = None
    groups = ()

    def authenticated_userid(self, request):
        return self.userid

    def effective_principals(self, request):
        from pyramid.security import Everyone
        from pyramid.security import Authenticated
        principals = [Everyone]
        userid = self.authenticated_userid(request)
        if userid:
            principals += [Authenticated]
            principals += [userid]
            principals += self.groups
        return principals

class RPCError(Exception):
    code = None
    message = None

    def __init__(self, code, message):
        self.code = code
        self.message = message

class SimpleRPCFixture(object):
    package = 'pyramid_rpc.tests.fixtures.simplerpc'
    path = None
    secure_path = None
    content_type = None

    def callFUT(self, path, method, body):
        """ Should raise RPCError if call fails."""
        return self.testapp.post(path,
                                 content_type=self.content_type,
                                 params=body)

    def setUp(self):
        from pyramid.config import Configurator
        from pyramid.authorization import ACLAuthorizationPolicy
        config = Configurator(package=self.package)
        config.include(self.package)
        self.auth_policy = DummyAuthenticationPolicy()
        config.set_authentication_policy(self.auth_policy)
        config.set_authorization_policy(ACLAuthorizationPolicy())
        app = config.make_wsgi_app()
        self.testapp = TestApp(app)
        self.config = config
        self.config.begin()

    def tearDown(self):
        self.config.end()

    def test_basic(self):
        result = self.callFUT(self.path, 'basic', ())
        self.assertEqual(result, 'basic')

    def test_exc(self):
        self.assertRaises(RPCError, self.callFUT, self.path, 'exc', ())

    def test_decorated_method(self):
        result = self.callFUT(self.path, 'create', (1, 2))
        self.assertEqual(result, 'create 1 2')

    def test_decorated_class_method(self):
        result = self.callFUT(self.path, 'class_hello', ('bob',))
        self.assertEqual(result, 'hello, bob')

    def test_basic_hello(self):
        result = self.callFUT(self.path, 'hello', ('bob',))
        self.assertEqual(result, 'hello, bob, stranger')

    def test_secure_hello_with_credentials(self):
        self.auth_policy.userid = 'bob'
        result = self.callFUT(self.secure_path, 'hello', ('bob',))
        self.assertEqual(result, 'hello, bob, friend')

    def test_secure_hello_with_no_credentials(self):
        self.assertRaises(RPCError, self.callFUT,
                          self.secure_path, 'hello', ('bob',))

class TestJSONRPC(SimpleRPCFixture, unittest.TestCase):
    path = '/api/jsonrpc'
    secure_path = '/api/jsonrpc/secure'
    content_type = 'application/json'

    def callFUT(self, path, method, params):
        body = {
            'id': 5,
            'jsonrpc': '2.0',
            'method': method,
            'params': params,
        }
        resp = super(TestJSONRPC, self).callFUT(path, method, json.dumps(body))
        self.assertEqual(resp.status_int, 200)
        self.assertEqual(resp.content_type, 'application/json')
        result = resp.json
        self.assertEqual(result['jsonrpc'], '2.0')
        self.assertEqual(result['id'], 5)
        if 'error' in result:
            raise RPCError(result['error']['code'], result['error']['message'])
        return result['result']

class TestXMLRPC(SimpleRPCFixture, unittest.TestCase):
    path = '/api/xmlrpc'
    secure_path = '/api/xmlrpc/secure'
    content_type = 'text/xml'

    def callFUT(self, path, method, params):
        body = xmlrpclib.dumps(params, methodname=method)
        resp = super(TestXMLRPC, self).callFUT(path, method, body)
        self.assertEqual(resp.status_int, 200)
        self.assertEqual(resp.content_type, 'text/xml')
        try:
            result = xmlrpclib.loads(resp.body)[0][0]
        except xmlrpclib.Fault as e:
            raise RPCError(e.faultCode, e.faultString)
        return result



import logging

import venusian
from pyramid.exceptions import ConfigurationError
from pyramid.httpexceptions import HTTPNotFound
from pyramid.security import NO_PERMISSION_REQUIRED

from .compat import (
    PY3,
    is_nonstr_iter,
    xmlrpclib,
)
from .mapper import MapplyViewMapper
from .mapper import ViewMapperArgsInvalid
from .util import combine


log = logging.getLogger(__name__)


_marker = object()


DEFAULT_RENDERER = "xmlrpc"


class XmlRpcError(xmlrpclib.Fault):
    faultCode = None
    faultString = None

    def __init__(self):
        xmlrpclib.Fault.__init__(self, self.faultCode, self.faultString)


class XmlRpcApplicationError(XmlRpcError):
    faultCode = -32500
    faultString = 'application error'


class XmlRpcMethodNotFound(XmlRpcError):
    faultCode = -32601
    faultString = 'server error; requested method not found'


class XmlRpcInvalidMethodParams(XmlRpcError):
    faultCode = -32602
    faultString = 'server error; invalid method params'


class XmlRpcParseError(XmlRpcError):
    faultCode = -32700
    faultString = 'parse error; not well formed'


class XMLRPCRenderer:
    def __init__(self, **kw):
        self.kw = kw

    def __call__(self, info):
        def _render(value, system):
            request = system.get('request')
            if request is not None:
                response = request.response
                ct = response.content_type
                if ct == response.default_content_type:
                    response.content_type = 'text/xml'

            if isinstance(value, xmlrpclib.Fault):
                obj = value
            else:
                obj = (value,)

            return xmlrpclib.dumps(obj, methodresponse=True, **self.kw)
        return _render


def exception_view(exc, request):
    if isinstance(exc, xmlrpclib.Fault):
        fault = exc
    elif isinstance(exc, HTTPNotFound):
        fault = XmlRpcMethodNotFound()
        log.debug('xml-rpc method not found "%s"', request.rpc_method)
    elif isinstance(exc, ViewMapperArgsInvalid):
        fault = XmlRpcInvalidMethodParams()
        log.debug('xml-rpc method not found "%s"', request.rpc_method)
    else:
        fault = XmlRpcApplicationError()
        log.exception('xml-rpc exception "%s"', exc)

    return fault


class EndpointPredicate(object):
    def __init__(self, val, config):
        self.val = val

    def text(self):
        return 'jsonrpc endpoint = %s' % self.val

    phash = text

    def __call__(self, info, request):
        if self.val:
            # find the endpoint info
            key = info['route'].name
            endpoint = request.registry.xmlrpc_endpoints[key]

            # parse the request body
            setup_request(endpoint, request)

            # update request with endpoint information
            request.rpc_endpoint = endpoint

            # Always return True so that even if it isn't a valid RPC it
            # will fall through to the notfound_view which will still
            # return a valid XML-RPC response.
            return True


class MethodPredicate(object):
    def __init__(self, val, config):
        self.method = val

    def text(self):
        return 'xmlrpc method = %s' % self.method

    phash = text

    def __call__(self, context, request):
        return getattr(request, 'rpc_method') == self.method


class Endpoint(object):
    def __init__(self, name, default_mapper, default_renderer):
        self.name = name
        self.default_mapper = default_mapper
        self.default_renderer = default_renderer


def setup_request(endpoint, request):
    try:
        params, method = xmlrpclib.loads(request.body)
    except Exception:
        raise XmlRpcParseError

    request.rpc_args = params
    request.rpc_method = method

    if method is None:
        raise XmlRpcMethodNotFound


def add_xmlrpc_endpoint(config, name, *args, **kw):
    """Add an endpoint for handling XML-RPC.

    ``name``

        The name of the endpoint.

    ``default_mapper``

        A default view mapper that will be passed as the ``mapper``
        argument to each of the endpoint's methods.

    A XML-RPC method also accepts all of the arguments supplied to
    Pyramid's ``add_route`` method.

    """
    default_mapper = kw.pop('default_mapper', MapplyViewMapper)
    default_renderer = kw.pop('default_renderer', DEFAULT_RENDERER)

    endpoint = Endpoint(
        name,
        default_mapper=default_mapper,
        default_renderer=default_renderer,
    )

    config.registry.xmlrpc_endpoints[name] = endpoint
    kw['xmlrpc_endpoint'] = True

    config.add_route(name, *args, **kw)
    config.add_view(exception_view, route_name=name, context=Exception,
                    permission=NO_PERMISSION_REQUIRED,
                    renderer=endpoint.default_renderer)


def add_xmlrpc_method(config, view, **kw):
    """Add a method to a XML-RPC endpoint.

    ``endpoint``

        The name of the endpoint.

    ``method``

        The name of the method.

    A XML-RPC method also accepts all of the arguments supplied to
    Pyramid's ``add_view`` method.

    A view mapper is registered by default which will match the
    ``request.rpc_args`` to parameters on the view. To override this
    behavior simply set the ``mapper`` argument to None or another
    view mapper.

    .. note::

       An endpoint **must** be defined before methods may be added.

    """
    endpoint_name = kw.pop('endpoint', kw.pop('route_name', None))
    if endpoint_name is None:
        raise ConfigurationError(
            'Cannot register a XML-RPC endpoint without specifying the '
            'name of the endpoint.')

    endpoint = config.registry.xmlrpc_endpoints.get(endpoint_name)
    if endpoint is None:
        raise ConfigurationError(
            'Could not find an endpoint with the name "%s".' % endpoint_name)

    method = kw.pop('method', None)
    if method is None:
        raise ConfigurationError(
            'Cannot register a XML-RPC method without specifying the '
            '"method"')

    mapper = kw.pop('mapper', _marker)
    if mapper is _marker:
        # only override mapper if not supplied
        mapper = endpoint.default_mapper
    kw['mapper'] = mapper

    renderer = kw.pop('renderer', _marker)
    if renderer is _marker:
        # Only override renderer if not supplied
        renderer = endpoint.default_renderer
    kw['renderer'] = renderer

    kw['xmlrpc_method'] = method

    config.add_view(view, route_name=endpoint_name, **kw)


class xmlrpc_method(object):
    """This decorator may be used with pyramid view callables to enable
    them to respond to XML-RPC method calls.

    If ``method`` is not supplied, then the callable name will be used
    for the method name.

    ``_depth`` may be specified when wrapping ``xmlrpc_method`` in another
    decorator. The value should reflect how many stack frames are between
    the wrapped target and ``xmlrpc_method``. Thus a decorator one level deep
    would pass in ``_depth=1``.

    This is the lazy analog to the
    :func:`~pyramid_rpc.xmlrpc.add_xmlrpc_method`` and accepts all of
    the same arguments.

    """
    def __init__(self, method=None, **kw):
        self.method = method
        self.kw = kw

    def __call__(self, wrapped):
        kw = self.kw.copy()
        kw['method'] = self.method or wrapped.__name__
        depth = kw.pop('_depth', 0)

        def callback(context, name, ob):
            config = context.config.with_package(info.module)
            config.add_xmlrpc_method(view=ob, **kw)

        info = venusian.attach(wrapped, callback, category='pyramid',
                               depth=depth + 1)
        if info.scope == 'class':
            # ensure that attr is set if decorating a class method
            kw.setdefault('attr', wrapped.__name__)

        kw['_info'] = info.codeinfo # fbo action_method
        return wrapped


def includeme(config):
    """ Set up standard configurator registrations.  Use via:

    .. code-block:: python

       config = Configurator()
       config.include('pyramid_rpc.xmlrpc')

    Once this function has been invoked, two new directives will be
    available on the configurator:

    - ``add_xmlrpc_endpoint``: Add an endpoint for handling XML-RPC.

    - ``add_xmlrpc_method``: Add a method to a XML-RPC endpoint.

    """
    if not hasattr(config.registry, 'xmlrpc_endpoints'):
        config.registry.xmlrpc_endpoints = {}

    config.add_view_predicate('xmlrpc_method', MethodPredicate)
    config.add_route_predicate('xmlrpc_endpoint', EndpointPredicate)

    config.add_directive('add_xmlrpc_endpoint', add_xmlrpc_endpoint)
    config.add_directive('add_xmlrpc_method', add_xmlrpc_method)
    config.add_renderer(name=DEFAULT_RENDERER, factory=XMLRPCRenderer())
    config.add_view(exception_view, context=xmlrpclib.Fault,
                    permission=NO_PERMISSION_REQUIRED,
                    renderer=DEFAULT_RENDERER)



"""Pyramid View Gateway implementation.

A :class:`~pyramid_rpc.amfgateway.PyramidGateway` instance implements a
Pyramid view compatible callable object, allowing it to be configured like any
other Pyramid view.

"""
import pyamf
from pyamf import remoting
from pyamf.remoting import gateway

from pyramid.httpexceptions import (
    HTTPBadRequest,
    HTTPInternalServerError,
    HTTPMethodNotAllowed,
)
from pyramid.response import Response


class PyramidGateway(gateway.BaseGateway):
    """Pyramid View Remoting Gateway

    :param expose_request:
        Defaults to True. Whether or not the service function should be
        called with the Pyramid request object.
    :param debug: Boolean toggling debug mode.

    """

    def __init__(self, *args, **kwargs):
        kwargs['expose_request'] = kwargs.get('expose_request', True)
        gateway.BaseGateway.__init__(self, *args, **kwargs)

    def getResponse(self, request, amf_request):
        """Process the AMF request, returning an AMF response"""
        response = remoting.Envelope(amf_request.amfVersion)

        for name, message in amf_request:
            request.amf_request = message

            processor = self.getProcessor(message)
            response[name] = processor(message, http_request=request)
        return response

    def __call__(self, request):
        """Processes and dispatches the request"""
        if request.method != 'POST':
            return HTTPMethodNotAllowed(['POST'])

        body = request.body
        stream = None
        timezone_offset = self._get_timezone_offset()

        # Decode the request
        try:
            amf_request = remoting.decode(body, strict=self.strict,
                                          logger=self.logger,
                                          timezone_offset=timezone_offset)
        except (pyamf.DecodeError, IOError):
            if self.logger:
                self.logger.exception('Error decoding AMF request')

            response = "400 Bad Request\n\nThe request body was unable to " \
                "be successfully decoded."

            if self.debug:
                response += "\n\nTraceback:\n\n%s" % gateway.format_exception()
            return HTTPBadRequest(detail=response)
        except Exception:
            if self.logger:
                self.logger.exception('Unexpected error decoding AMF request')

            response = ("500 Internal Server Error\n\nAn unexpected error "
                "occurred whilst decoding.")

            if self.debug:
                response += "\n\nTraceback:\n\n%s" % gateway.format_exception()
            return HTTPInternalServerError(detail=response)

        if self.logger:
            self.logger.debug("AMF Request: %r" % amf_request)

        # Process the request
        try:
            response = self.getResponse(request, amf_request)
        except Exception:
            if self.logger:
                self.logger.exception('Error processing AMF request')

            response = ("500 Internal Server Error\n\nThe request was "
                "unable to be successfully processed.")

            if self.debug:
                response += "\n\nTraceback:\n\n%s" % gateway.format_exception()

            return HTTPInternalServerError(detail=response)

        if self.logger:
            self.logger.debug("AMF Response: %r" % response)

        # Encode the response
        try:
            stream = remoting.encode(response, strict=self.strict,
                timezone_offset=timezone_offset)
        except:
            if self.logger:
                self.logger.exception('Error encoding AMF request')

            response = ("500 Internal Server Error\n\nThe request was "
                "unable to be encoded.")

            if self.debug:
                response += "\n\nTraceback:\n\n%s" % gateway.format_exception()

            return HTTPInternalServerError(detail=response)

        buf = stream.getvalue()

        http_response = Response(content_type=remoting.CONTENT_TYPE)
        http_response.headers['Server'] = gateway.SERVER_NAME
        http_response.write(buf)
        return http_response



# stole from pyramid 1.4
def combine(*decorators):
    def decorated(view_callable):
        # reversed() is allows a more natural ordering in the api
        for decorator in reversed(decorators):
            view_callable = decorator(view_callable)
        return view_callable
    return decorated



import json
import unittest

from pyramid import testing

from webtest import TestApp


class Test_add_jsonrpc_method(unittest.TestCase):

    def setUp(self):
        self.config = testing.setUp()
        self.config.include('pyramid_rpc.jsonrpc')

    def tearDown(self):
        testing.tearDown()

    def test_with_undefined_endpoint(self):
        from pyramid.exceptions import ConfigurationError
        config = self.config
        self.assertRaises(ConfigurationError,
                          config.add_jsonrpc_method,
                          lambda r: None, endpoint='rpc', method='foo')

    def test_with_missing_endpoint_param(self):
        from pyramid.exceptions import ConfigurationError
        config = self.config
        config.add_jsonrpc_endpoint('rpc', '/api/jsonrpc')
        self.assertRaises(ConfigurationError,
                          config.add_jsonrpc_method,
                          lambda r: None, method='dummy')

    def test_with_no_method_param(self):
        from pyramid.exceptions import ConfigurationError
        config = self.config
        config.add_jsonrpc_endpoint('rpc', '/api/jsonrpc')
        self.assertRaises(ConfigurationError,
                          config.add_jsonrpc_method,
                          lambda r: None, endpoint='rpc')


class TestJSONRPCIntegration(unittest.TestCase):

    def setUp(self):
        self.config = testing.setUp()

    def tearDown(self):
        testing.tearDown()

    def _callFUT(self, app, method, params, id=5, version='2.0',
                 path='/api/jsonrpc', content_type='application/json',
                 expect_error=False):
        body = {}
        if id is not None:
            body['id'] = id
        if version is not None:
            body['jsonrpc'] = version
        if method is not None:
            body['method'] = method
        if params is not None:
            body['params'] = params
        resp = app.post(path, content_type=content_type,
                        params=json.dumps(body))
        self.assertEqual(resp.status_int, 200)
        if id is not None or expect_error:
            self.assertEqual(resp.content_type, 'application/json')
            result = resp.json
            self.assertEqual(result['jsonrpc'], '2.0')
            self.assertEqual(result['id'], id)
        else:
            result = resp.json
            self.assertEqual(result, '')
        return result

    def test_it(self):
        def view(request, a, b):
            return {'a': a, 'b': b}
        config = self.config
        config.include('pyramid_rpc.jsonrpc')
        config.add_jsonrpc_endpoint('rpc', '/api/jsonrpc')
        config.add_jsonrpc_method(view, endpoint='rpc', method='dummy')
        app = config.make_wsgi_app()
        app = TestApp(app)
        result = self._callFUT(app, 'dummy', [2, 3])
        self.assertEqual(result['result'], {'a': 2, 'b': 3})

    def test_it_with_no_mapper(self):
        def view(request):
            return request.rpc_args[0]
        config = self.config
        config.include('pyramid_rpc.jsonrpc')
        config.add_jsonrpc_endpoint('rpc', '/api/jsonrpc')
        config.add_jsonrpc_method(view, endpoint='rpc', method='dummy',
                                  mapper=None)
        app = config.make_wsgi_app()
        app = TestApp(app)
        result = self._callFUT(app, 'dummy', [2, 3])
        self.assertEqual(result['result'], 2)

    def test_it_with_multiple_methods(self):
        def view(request, a, b):
            return a
        config = self.config
        config.include('pyramid_rpc.jsonrpc')
        config.add_jsonrpc_endpoint('rpc', '/api/jsonrpc')
        config.add_jsonrpc_method(view, endpoint='rpc', method='dummy')
        config.add_jsonrpc_method(lambda r: 'fail',
                                  endpoint='rpc', method='dummy2')
        app = config.make_wsgi_app()
        app = TestApp(app)
        result = self._callFUT(app, 'dummy', [2, 3])
        self.assertEqual(result['result'], 2)

    def test_it_with_batched_requests(self):
        def view(request, a, b):
            return [a, b]
        config = self.config
        config.include('pyramid_rpc.jsonrpc')
        config.add_jsonrpc_endpoint('rpc', '/api/jsonrpc')
        config.add_jsonrpc_method(view, endpoint='rpc', method='dummy')
        app = config.make_wsgi_app()
        app = TestApp(app)
        body = [
            {'id': 1, 'jsonrpc': '2.0', 'method': 'dummy', 'params': [2, 3]},
            {'id': 2, 'jsonrpc': '2.0', 'method': 'dummy', 'params': {'a': 3, 'b': 2}},
        ]
        resp = app.post('/api/jsonrpc', content_type='application/json',
                        params=json.dumps(body))
        self.assertEqual(resp.status_int, 200)
        result = resp.json
        result1 = [r for r in result if r['id'] == 1][0]
        result2 = [r for r in result if r['id'] == 2][0]
        self.assertEqual(result1, {'id': 1, 'jsonrpc': '2.0', 'result': [2, 3]})
        self.assertEqual(result2, {'id': 2, 'jsonrpc': '2.0', 'result': [3, 2]})

    def test_it_with_no_version(self):
        config = self.config
        config.include('pyramid_rpc.jsonrpc')
        config.add_jsonrpc_endpoint('rpc', '/api/jsonrpc')
        app = config.make_wsgi_app()
        app = TestApp(app)
        result = self._callFUT(app, 'dummy', [2, 3], version=None)
        self.assertEqual(result['error']['code'], -32600)

    def test_it_with_no_method(self):
        config = self.config
        config.include('pyramid_rpc.jsonrpc')
        config.add_jsonrpc_endpoint('rpc', '/api/jsonrpc')
        app = config.make_wsgi_app()
        app = TestApp(app)
        result = self._callFUT(app, None, [2, 3])
        self.assertEqual(result['error']['code'], -32600)

    def test_it_with_no_id(self):
        def view(request, a, b):
            return a
        config = self.config
        config.include('pyramid_rpc.jsonrpc')
        config.add_jsonrpc_endpoint('rpc', '/api/jsonrpc')
        config.add_jsonrpc_method(view, endpoint='rpc', method='dummy')
        app = config.make_wsgi_app()
        app = TestApp(app)
        result = self._callFUT(app, 'dummy', [2, 3], id=None)
        self.assertEqual(result, '')

    def test_it_with_batched_notifications(self):
        # a notification is a request with no id
        def view(request, a, b):
            return [a, b]
        config = self.config
        config.include('pyramid_rpc.jsonrpc')
        config.add_jsonrpc_endpoint('rpc', '/api/jsonrpc')
        config.add_jsonrpc_method(view, endpoint='rpc', method='dummy')
        app = config.make_wsgi_app()
        app = TestApp(app)
        body = [
            {'jsonrpc': '2.0', 'method': 'dummy', 'params': [2, 3]},
            {'jsonrpc': '2.0', 'method': 'dummy', 'params': {'a': 3, 'b': 2}},
        ]
        resp = app.post('/api/jsonrpc', content_type='application/json',
                        params=json.dumps(body))
        self.assertEqual(resp.status_int, 200)
        self.assertEqual(resp.body, b'')

    def test_it_with_no_params(self):
        def view(request):
            self.assertEqual(request.rpc_args, ())
            return 'no params'
        config = self.config
        config.include('pyramid_rpc.jsonrpc')
        config.add_jsonrpc_endpoint('rpc', '/api/jsonrpc')
        config.add_jsonrpc_method(view, endpoint='rpc', method='dummy')
        app = config.make_wsgi_app()
        app = TestApp(app)
        result = self._callFUT(app, 'dummy', None)
        self.assertEqual(result['result'], 'no params')

    def test_it_with_named_params(self):
        def view(request, three, four, five):
            self.assertEqual('%s%s%s'%(three,four,five), '345')
            return 'named params'
        config = self.config
        config.include('pyramid_rpc.jsonrpc')
        config.add_jsonrpc_endpoint('rpc', '/api/jsonrpc')
        config.add_jsonrpc_method(view, endpoint='rpc', method='dummy')
        app = config.make_wsgi_app()
        app = TestApp(app)
        result = self._callFUT(app, 'dummy', {'four':4, 'five':5, 'three':3})
        self.assertEqual(result['result'], 'named params')

    def test_it_with_named_params_and_default_values(self):
        def view(request, three, four = 4 , five = 'foo' ):
            self.assertEqual('%s%s%s'%(three,four,five), '345')
            return 'named params'
        config = self.config
        config.include('pyramid_rpc.jsonrpc')
        config.add_jsonrpc_endpoint('rpc', '/api/jsonrpc')
        config.add_jsonrpc_method(view, endpoint='rpc', method='dummy')
        app = config.make_wsgi_app()
        app = TestApp(app)
        result = self._callFUT(app, 'dummy', { 'five':5, 'three':3})
        self.assertEqual(result['result'], 'named params')

    def test_it_with_invalid_method(self):
        config = self.config
        config.include('pyramid_rpc.jsonrpc')
        config.add_jsonrpc_endpoint('rpc', '/api/jsonrpc')
        app = config.make_wsgi_app()
        app = TestApp(app)
        result = self._callFUT(app, 'foo', [2, 3])
        self.assertEqual(result['error']['code'], -32601)

    def test_it_with_invalid_body(self):
        config = self.config
        config.include('pyramid_rpc.jsonrpc')
        config.add_jsonrpc_endpoint('rpc', '/api/jsonrpc')
        app = config.make_wsgi_app()
        app = TestApp(app)
        resp = app.post('/api/jsonrpc', content_type='application/json',
                        params='{')
        self.assertEqual(resp.status_int, 200)
        self.assertEqual(resp.content_type, 'application/json')
        result = resp.json
        self.assertEqual(result['jsonrpc'], '2.0')
        self.assertEqual(result['error']['code'], -32700)

    def test_it_with_general_exception(self):
        def view(request, a, b):
            raise Exception
        config = self.config
        config.include('pyramid_rpc.jsonrpc')
        config.add_jsonrpc_endpoint('rpc', '/api/jsonrpc')
        config.add_jsonrpc_method(view, endpoint='rpc', method='dummy')
        app = config.make_wsgi_app()
        app = TestApp(app)
        result = self._callFUT(app, 'dummy', [2, 3])
        self.assertEqual(result['error']['code'], -32603)

    def test_it_with_rpc_error(self):
        from pyramid_rpc.jsonrpc import JsonRpcError
        def view(request):
            raise JsonRpcError(code=500, message='dummy')
        config = self.config
        config.include('pyramid_rpc.jsonrpc')
        config.add_jsonrpc_endpoint('rpc', '/api/jsonrpc')
        config.add_jsonrpc_method(view, endpoint='rpc', method='dummy')
        app = config.make_wsgi_app()
        app = TestApp(app)
        result = self._callFUT(app, 'dummy', [])
        self.assertEqual(result['error']['code'], 500)
        self.assertEqual(result['error']['message'], 'dummy')
        self.assertFalse('data' in result['error'])

    def test_it_with_rpc_error_with_data(self):
        from pyramid_rpc.jsonrpc import JsonRpcError
        def view(request):
            raise JsonRpcError(code=500, message='dummy', data='foo')
        config = self.config
        config.include('pyramid_rpc.jsonrpc')
        config.add_jsonrpc_endpoint('rpc', '/api/jsonrpc')
        config.add_jsonrpc_method(view, endpoint='rpc', method='dummy')
        app = config.make_wsgi_app()
        app = TestApp(app)
        result = self._callFUT(app, 'dummy', [])
        self.assertEqual(result['error']['code'], 500)
        self.assertEqual(result['error']['message'], 'dummy')
        self.assertEqual(result['error']['data'], 'foo')

    def test_it_with_cls_view(self):
        class view(object):
            def __init__(self, request):
                self.request = request

            def foo(self, a, b):
                return [a, b]
        config = self.config
        config.include('pyramid_rpc.jsonrpc')
        config.add_jsonrpc_endpoint('rpc', '/api/jsonrpc')
        config.add_jsonrpc_method(view, endpoint='rpc', method='dummy',
                                  attr='foo')
        app = config.make_wsgi_app()
        app = TestApp(app)
        result = self._callFUT(app, 'dummy', [2, 3])
        self.assertEqual(result['result'], [2, 3])

    def test_it_with_named_args_and_cls_view(self):
        class view(object):
            def __init__(self, request):
                self.request = request

            def foo(self, a, b):
                return [a, b]
        config = self.config
        config.include('pyramid_rpc.jsonrpc')
        config.add_jsonrpc_endpoint('rpc', '/api/jsonrpc')
        config.add_jsonrpc_method(view, endpoint='rpc', method='dummy',
                                  attr='foo')
        app = config.make_wsgi_app()
        app = TestApp(app)
        result = self._callFUT(app, 'dummy', {'b':3, 'a':2})
        self.assertEqual(result['result'], [2, 3])

    def test_it_with_default_args(self):
        def view(request, a, b, c='bar'):
            return [a, b, c]
        config = self.config
        config.include('pyramid_rpc.jsonrpc')
        config.add_jsonrpc_endpoint('rpc', '/api/jsonrpc')
        config.add_jsonrpc_method(view, endpoint='rpc', method='dummy')
        app = config.make_wsgi_app()
        app = TestApp(app)
        result = self._callFUT(app, 'dummy', [2, 3])
        self.assertEqual(result['result'], [2, 3, 'bar'])

    def test_it_with_missing_args(self):
        config = self.config
        config.include('pyramid_rpc.jsonrpc')
        config.add_jsonrpc_endpoint('rpc', '/api/jsonrpc')
        config.add_jsonrpc_method(lambda r, a, b: a,
                                  endpoint='rpc', method='dummy')
        app = config.make_wsgi_app()
        app = TestApp(app)
        result = self._callFUT(app, 'dummy', [2])
        self.assertEqual(result['error']['code'], -32602)

    def test_it_with_too_many_args(self):
        config = self.config
        config.include('pyramid_rpc.jsonrpc')
        config.add_jsonrpc_endpoint('rpc', '/api/jsonrpc')
        config.add_jsonrpc_method(lambda r, a, b: a,
                                  endpoint='rpc', method='dummy')
        app = config.make_wsgi_app()
        app = TestApp(app)
        result = self._callFUT(app, 'dummy', [2, 3, 4])
        self.assertEqual(result['error']['code'], -32602)

    def test_it_error_with_no_id(self):
        def view(request):
            raise Exception
        config = self.config
        config.include('pyramid_rpc.jsonrpc')
        config.add_jsonrpc_endpoint('rpc', '/api/jsonrpc')
        config.add_jsonrpc_method(view, endpoint='rpc', method='dummy')
        app = config.make_wsgi_app()
        app = TestApp(app)
        result = self._callFUT(app, 'dummy', [], id=None, expect_error=True)
        self.assertEqual(result['error']['code'], -32603)

    def test_it_with_decorator(self):
        def view(request):
            return 'foo'
        config = self.config
        config.include('pyramid_rpc.jsonrpc')
        config.add_jsonrpc_endpoint('rpc', '/api/jsonrpc')
        dummy_decorator = DummyDecorator()
        config.add_jsonrpc_method(view, endpoint='rpc', method='dummy',
                                  decorator=dummy_decorator)
        app = config.make_wsgi_app()
        app = TestApp(app)
        result = self._callFUT(app, 'dummy', [])
        self.assertEqual(result['result'], 'foo')
        self.assertTrue(dummy_decorator.called)

    def test_it_with_default_mapper(self):
        def view(request):
            return request.rpc_args
        config = self.config
        config.include('pyramid_rpc.jsonrpc')
        config.add_jsonrpc_endpoint('rpc', '/api/jsonrpc', default_mapper=None)
        config.add_jsonrpc_method(view, endpoint='rpc', method='dummy')
        app = config.make_wsgi_app()
        app = TestApp(app)
        result = self._callFUT(app, 'dummy', ['a', 'b', 'c'])
        self.assertEqual(result['result'], ['a', 'b', 'c'])

    def test_override_default_mapper(self):
        from pyramid_rpc.mapper import MapplyViewMapper
        def view(request, a, b, c):
            return (a, b, c)
        config = self.config
        config.include('pyramid_rpc.jsonrpc')
        config.add_jsonrpc_endpoint('rpc', '/api/jsonrpc', default_mapper=None)
        config.add_jsonrpc_method(view, endpoint='rpc', method='dummy',
                                  mapper=MapplyViewMapper)
        app = config.make_wsgi_app()
        app = TestApp(app)
        result = self._callFUT(app, 'dummy', ['a', 'b', 'c'])
        self.assertEqual(result['result'], ['a', 'b', 'c'])

    def test_it_with_default_renderer(self):
        def view(request):
            return 'bar'
        config = self.config
        config.include('pyramid_rpc.jsonrpc')
        dummy_renderer = DummyRenderer('foo')
        config.add_renderer('jsonrpc', dummy_renderer)
        config.add_jsonrpc_endpoint('rpc', '/api/jsonrpc',
                                    default_renderer='jsonrpc')
        config.add_jsonrpc_method(view, endpoint='rpc', method='dummy')
        app = config.make_wsgi_app()
        app = TestApp(app)
        result = self._callFUT(app, 'dummy', [])
        self.assertEqual(result['result'], 'foo')
        self.assertEqual(dummy_renderer.called, True)

    def test_override_default_renderer(self):
        def view(request):
            return 'bar'
        config = self.config
        config.include('pyramid_rpc.jsonrpc')
        dummy_renderer = DummyRenderer('foo')
        dummy_renderer2 = DummyRenderer('baz')
        config.add_renderer('jsonrpc', dummy_renderer)
        config.add_renderer('jsonrpc2', dummy_renderer2)
        config.add_jsonrpc_endpoint('rpc', '/api/jsonrpc',
                                    default_renderer='jsonrpc')
        config.add_jsonrpc_method(view, endpoint='rpc', method='dummy',
                                  renderer='jsonrpc2')
        app = config.make_wsgi_app()
        app = TestApp(app)
        result = self._callFUT(app, 'dummy', [])
        self.assertEqual(result['result'], 'baz')
        self.assertEqual(dummy_renderer.called, False)
        self.assertEqual(dummy_renderer2.called, True)

    def test_nonascii_request(self):
        def view(request, a):
            return a
        config = self.config
        config.include('pyramid_rpc.jsonrpc')
        config.add_jsonrpc_endpoint('rpc', '/api/jsonrpc')
        config.add_jsonrpc_method(view, endpoint='rpc', method='dummy')
        app = config.make_wsgi_app()
        app = TestApp(app)
        val = b'S\xc3\xa9bastien'.decode('utf-8')
        result = self._callFUT(app, 'dummy', [val])
        self.assertEqual(result['result'], val)


class TestGET(unittest.TestCase):

    def setUp(self):
        def view(request, a, b=3):
            return [a, b]
        self.config = config = testing.setUp()
        config.include('pyramid_rpc.jsonrpc')
        config.add_jsonrpc_endpoint('rpc', '/api/jsonrpc')
        config.add_jsonrpc_method(view, endpoint='rpc', method='dummy')

    def tearDown(self):
        testing.tearDown()

    def _makeTestApp(self):
        app = self.config.make_wsgi_app()
        app = TestApp(app)
        return app

    def _callFUT(self, app, method, params, id='5', version='2.0',
                 path='/api/jsonrpc', expect_error=False):
        qitems = []
        if id is not None:
            qitems.append(('id', id))
        if version is not None:
            qitems.append(('jsonrpc', version))
        if method is not None:
            qitems.append(('method', method))
        if params is not None:
            if isinstance(params, str):
                qitems.append(('params', params))
            else:
                qitems.append(('params', json.dumps(params)))
        #import pdb; pdb.set_trace()
        resp = app.get(path, params=qitems)
        self.assertEqual(resp.status_int, 200)
        if id is not None or expect_error:
            self.assertEqual(resp.content_type, 'application/json')
            result = resp.json
            self.assertEqual(result['jsonrpc'], '2.0')
            self.assertEqual(result['id'], id)
        else:
            result = resp.json
            self.assertEqual(result, '')
        return result

    def test_it(self):
        app = self._makeTestApp()
        result = self._callFUT(app, 'dummy', [2, 3])
        self.assertEqual(result['result'], [2, 3])

    def test_it_named_args(self):
        app = self._makeTestApp()
        result = self._callFUT(app, 'dummy', {'a': 2, 'b': 3})
        self.assertEqual(result['result'], [2, 3])

    def test_it_with_default_args(self):
        app = self._makeTestApp()
        result = self._callFUT(app, 'dummy', [2])
        self.assertEqual(result['result'], [2, 3])

    def test_it_with_too_few_args(self):
        app = self._makeTestApp()
        result = self._callFUT(app, 'dummy', [])
        self.assertEqual(result['error']['code'], -32602)

    def test_it_with_too_many_args(self):
        app = self._makeTestApp()
        result = self._callFUT(app, 'dummy', [2, 3, 4])
        self.assertEqual(result['error']['code'], -32602)

    def test_it_with_unparseable_args(self):
        app = self._makeTestApp()
        result = self._callFUT(app, 'dummy', 'foo', id=None, expect_error=True)
        self.assertEqual(result['error']['code'], -32700)

    def test_it_with_missing_args(self):
        app = self._makeTestApp()
        result = self._callFUT(app, 'dummy', None)
        self.assertEqual(result['error']['code'], -32602)

    def test_it_with_no_id(self):
        app = self._makeTestApp()
        self._callFUT(app, 'dummy', [2, 3], id=None)

    def test_it_error_with_no_id(self):
        def view(request):
            raise Exception
        config = self.config
        config.add_jsonrpc_method(view, endpoint='rpc', method='err')
        app = self._makeTestApp()
        result = self._callFUT(app, 'err', [], id=None, expect_error=True)
        self.assertEqual(result['error']['code'], -32603)

    def test_PUT(self):
        app = self._makeTestApp()
        response = app.put('/api/jsonrpc')
        result = response.json_body
        self.assertEqual(result['error']['code'], -32600)

    def test_DELETE(self):
        app = self._makeTestApp()
        response = app.delete('/api/jsonrpc')
        result = response.json_body
        self.assertEqual(result['error']['code'], -32600)


class DummyDecorator(object):
    called = False

    def __call__(self, view):
        def wrapper(context, request):
            self.called = True
            return view(context, request)
        return wrapper


class DummyRenderer(object):
    called = False

    def __init__(self, result):
        self.result = result

    def __call__(self, info):
        def _render(value, system):
            self.called = True
            value['result'] = self.result
            return json.dumps(value)
        return _render



import inspect

from pyramid.interfaces import IViewMapperFactory

from zope.interface import implementer

from pyramid_rpc.compat import PY3

if PY3: # pragma: no cover
    def _inspect_ob(f):
        im = False

        if hasattr(f, '__func__'):
            im = True

        elif not hasattr(f, '__defaults__'):
            if hasattr(f, '__call__'):
                f = f.__call__
                if hasattr(f, '__func__'):
                    im = True

        if im:
            f = f.__func__
            c = f.__code__
            defaults = f.__defaults__
            names = c.co_varnames[1:c.co_argcount]
        else:
            c = f.__code__
            defaults = f.__defaults__
            names = c.co_varnames[:c.co_argcount]

        return names, defaults
else:
    def _inspect_ob(f):
        im = False

        if hasattr(f, 'im_func'):
            im = True

        elif not hasattr(f, 'func_defaults'):
            if hasattr(f, '__call__'):
                f = f.__call__
                if hasattr(f, 'im_func'):
                    im = True

        if im:
            f = f.im_func
            c = f.func_code
            defaults = f.func_defaults
            names = c.co_varnames[1:c.co_argcount]
        else:
            c = f.func_code
            defaults = f.func_defaults
            names = c.co_varnames[:c.co_argcount]

        return names, defaults

@implementer(IViewMapperFactory)
class MapplyViewMapper(object):

    def __init__(self, **kw):
        self.attr = kw.get('attr')

    def __call__(self, view):
        attr = self.attr
        if inspect.isclass(view):
            def _class_view(context, request):
                params = getattr(request, 'rpc_args', ())
                keywords = dict(request.params.items())
                if request.matchdict:
                    keywords.update(request.matchdict)
                if isinstance(params, dict):
                    keywords.update(params)
                    params = tuple()
                else:
                    params = tuple(params)
                if attr is None:
                    inst = view(request)
                    response = self.mapply(inst, params, keywords)
                else:
                    inst = view(request)
                    response = self.mapply(getattr(inst, attr), params,
                                           keywords)
                request.__view__ = inst
                return response
            mapped_view = _class_view
        else:
            def _nonclass_view(context, request):
                params = getattr(request, 'rpc_args', ())
                keywords = dict(request.params.items())
                if request.matchdict:
                    keywords.update(request.matchdict)
                if isinstance(params, dict):
                    keywords.update(params)
                    params = (request,)
                else:
                    params = (request,) + tuple(params)
                if attr is None:
                    response = self.mapply(view, params, keywords)
                else:
                    response = self.mapply(getattr(view, attr), params,
                                           keywords)
                return response
            mapped_view = _nonclass_view

        return mapped_view

    def mapply(self, ob, positional, keyword):
        names, defaults = _inspect_ob(ob)

        nargs = len(names)
        args = []
        if positional:
            positional = list(positional)
            if len(positional) > nargs:
                raise ViewMapperArgsInvalid('too many arguments')
            args = positional

        get = keyword.get
        nrequired = len(names) - (len(defaults or ()))
        for index in range(len(args), len(names)):
            name = names[index]
            v = get(name, args)
            if v is args:
                if index < nrequired:
                    raise ViewMapperArgsInvalid(
                        'argument %s was omitted' % name)
                else:
                    v = defaults[index - nrequired]
            args.append(v)

        args = tuple(args)
        return ob(*args)


class ViewMapperArgsInvalid(TypeError):
    pass



##############################################################################
#
# Copyright (c) 2008-2010 Agendaless Consulting and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the BSD-like license at
# http://www.repoze.org/LICENSE.txt.  A copy of the license should accompany
# this distribution.  THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL
# EXPRESS OR IMPLIED WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND
# FITNESS FOR A PARTICULAR PURPOSE
#
##############################################################################

import os
import sys

from setuptools import setup, find_packages

py_version = sys.version_info[:2]
PY3 = py_version[0] == 3

here = os.path.abspath(os.path.dirname(__file__))

try:
    README = open(os.path.join(here, 'README.rst')).read()
except:
    README = ''

install_requires = [
    'venusian>=1.0a7',
    'pyramid>=1.4',
]

tests_require = install_requires + [
    'WebTest',
]

testing_extras = tests_require + [
    'coverage',
    'nose',
]

docs_require = [
    'pylons-sphinx-themes',
    'Sphinx >= 1.3.1',
]

setup(name='pyramid_rpc',
      version='0.7',
      description='RPC support for the Pyramid web framework',
      long_description=README,
      classifiers=[
          "Intended Audience :: Developers",
          "Programming Language :: Python",
          "Programming Language :: Python :: 2",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 2.6",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3.2",
          "Programming Language :: Python :: 3.3",
          "Programming Language :: Python :: 3.4",
          "Framework :: Pyramid",
      ],
      keywords='web wsgi pyramid pylons xml-rpc json-rpc',
      author="Ben Bangert",
      author_email="ben@groovie.org",
      maintainer='Michael Merickel',
      maintainer_email='michael@merickel.org',
      url='http://docs.pylonsproject.org/projects/pyramid_rpc/en/latest/',
      license="BSD-derived (http://www.repoze.org/LICENSE.txt)",
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      tests_require=tests_require,
      install_requires=install_requires,
      extras_require={
          'testing': testing_extras,
          'docs': docs_require,
          'amf': ['pyamf<=0.6.99'],
      },
      test_suite="pyramid_rpc.tests",
      )
